//***********************************************************************
//
//
//                              HFPIN.CPP
//
//
// Author(s) : Serge GUILLAUME, Brigitte CHARNOMORDIC 
// FISPRO Version : 3.2 - Copyright INRA - Cemagref - 2002 - IDDN.FR.001.030024.000.R.P.2005.003.31235
// Licence http://www.cecill.info/licences/Licence_CeCILL_V2-en.html/
// Last modification date:  July 31, 2009 - Contact : fispro@supagro.inra.fr
// File : INHFP class functions used by FISPRO, 
//        part of package hfp (Hierarchical fuzzy partitioning)
//        requires library fisprob
//
//**********************************************************************

#include "hfpin.h"

#define NOT_INIT 147.852369

INHFP::INHFP(double *t, int n, int hier, int nmf, double tol)
  //********************************************************* 
  : FISIN(0)
{
  int nb;
 
  Init();

  HierType = hier;

  ValInf = NOT_INIT;
  ValSup = NOT_INIT;

  NmfMax = ((nmf < MAX_MF) ? nmf : MAX_MF);

  InitCont(t, n, tol);

  switch(HierType)
    {
    case T_HFP    : InMerge(NULL); break;
    case T_KMEANS : HKmeans();  break;
    case T_REG    : HRegular(); break;
    default       : HRegular(); break;
    }

  nb = ((Nmf < Nvertices) ? Nmf : Nvertices);
  MfInstall(nb, true); // Vertices are used as MF centres

}

INHFP::INHFP(ifstream & f, int bufsize, int num, char * s_hier, int DistSymb, int crit, 
	     double Penal, int NSDist, int NSVar, int SimpDistfie, int ng) 
  //**********************************************************************
  : FISIN(0)
{
  char * tmp=NULL,  * buf=NULL;
  double Tab[2];

  tmp = new char[bufsize];
  buf = new char[bufsize];

  Init();

  do{                           // Skip empty lines
    f.getline(buf, bufsize);
  }while((strlen(buf) == 0) || (buf[0]==0x0D));  // Ignore ^M (DOS file end of line)


  sprintf(tmp,"Active=");              
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }

  if(SearchStr(buf, tmp))
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
      throw std::runtime_error( ErrorMsg );
    }
  if( ! strncmp(tmp, "no", 4) )  active = false;   
  else  if( ! strncmp(tmp, "yes", 4) ) active = true;
  else 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: yes or no\n~ReadString~: %.50s~", tmp);
      throw std::runtime_error( ErrorMsg );
    } 

  f.getline(buf, bufsize);              // Name
  sprintf(tmp,"Name=");
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }
  if(SearchStr(buf, tmp)) 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~StringSeparatorNotFoundInString~: %.50s~", buf );
      throw std::runtime_error( ErrorMsg );
    }
  SetName(tmp);

  f.getline(buf, bufsize);             
  sprintf(tmp,"Range=");
  if( strncmp(tmp, buf, strlen(tmp)) )
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }

  if((strlen(buf + strlen(tmp)) == 0) || (* (buf + strlen(tmp))==0x0D))
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~~EmptyString~: %.50s~", tmp);
      throw std::runtime_error( ErrorMsg );
    }

  if((SearchNb(buf, Tab, 2, SEPARE, START_NB, END_NB)) != 2) 
    {
      ValInf = FisMknan();
      ValSup = FisMknan(); 
    }
  else 
    {
      ValInf = Tab[0];
      ValSup = Tab[1];
    }

  f.getline(buf, bufsize);             
  sprintf(tmp,"NMFs=");
  if( strncmp(tmp, buf, strlen(tmp)) ) 
    {
      sprintf( ErrorMsg, "~ErrorInFISFile~\n~ExpectedString~: %.50s\n~ReadString~: %.50s~", tmp, buf );
      throw std::runtime_error( ErrorMsg );
    }
  NmfMax = atoi(buf + strlen(tmp));
  if(NmfMax > MAX_MF) NmfMax = MAX_MF;
  Symbolic = DistSymb;
  VarCrit = crit;       // Variance or distance
  Penalty = 1.0 + Penal;
  VarThres = 0.0;       // Will be set after ouitput definition
  NmfDist = NSDist;
  NmfVar = NSVar;
  Dsimpli = SimpDistfie;
  SetHierType(s_hier);

  delete [] tmp;    
  delete [] buf;
} 

void  INHFP::InitCont(double * T, int nb, double thres)
  //***************************************************
{
  if(! active) return;

  int i, j, cut;
  double diff;
  double * reg;

  reg = NULL;
  cut = false;
  j = 0;
  
  MinObserve = MaxObserve = T[0]; 
  
  for(i = 0; i < nb; i++)
    {
      if(T[i]< MinObserve) MinObserve = T[i];
      else if(T[i] > MaxObserve) MaxObserve = T[i];
    }

  cut = false;
  if(ValInf == NOT_INIT) ValInf = MinObserve;
  else if(ValInf > MinObserve) 
    {
      MinObserve = ValInf;
      cut = true;
    }
  if(ValSup == NOT_INIT) ValSup = MaxObserve;
  else if(ValSup < MaxObserve) 
    {
      MaxObserve = ValSup;
      cut = true;
    }

  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;

  if(cut)
    for(i = 0; i < nb; i++)
      if((T[i] < ValInf) || (T[i] > ValSup)) j++;
  // j is the number of values to be cut

  Val = new double [nb - j];
  Size = nb - j;  
  j = 0;    

  for(i = 0; i < nb; i++)
    {
      if((T[i] < ValInf) || (T[i] > ValSup)) continue;
      Val[j++] = (T[i] - ValInf) / diff;
    }

 if(thres > 1.)  KmeansInitPart((int) floor(thres));
  else UniqInitPart(thres);
  
  // Generate missing  values 
  for(i = 0; i < Size; i++)
    if(FisIsnan(Val[i]))  Val[i] = FisRand();

  Dist = new double * [NbUnique];
  for(i = 0; i < NbUnique; i++) Dist[i] = new double [NbUnique];
  
  if(HierType == T_REG) Nmf = NmfMax; // No limitation
  else if (HierType == T_KMEANS) 
    Nmf = ((NbUnique < NmfMax) ? NbUnique : NmfMax);
  else Nmf = NbUnique;

  NmfVar = ((Nmf < NmfVar) ? Nmf : NmfVar);
  Nvertices = ((NmfMax < Nmf) ? NmfMax : Nmf);

  //Mfdeg = new double [Nmf];
  
  Vertices = new TVertex * [Nvertices];
  for(i = 0; i < Nvertices; i++) 
    Vertices[i] = new TVertex  [i + 2];

  if(HierType == T_REG) 
    {
      reg = new double [Nvertices];
      if(Nvertices > 1)
	for(i = 0; i < Nvertices; i++)
	  reg[i] = (double) (i-1)/(Nvertices-1);
      else reg[0] = ValInf;
    }

  switch(HierType)
    {
    case T_HFP    : InitMf(Unique); break;
    case T_KMEANS : InitMf(Unique); break;
    case T_REG    : InitMf(reg); break;
    default       : InitMf(reg); break;
    }

  if(HierType == T_REG) delete [] reg;
} 

void INHFP::UniqInitPart(double equal)
  //**********************************
{
  if(equal < EPSILON) equal = EPSILON;
 
  SortUniq(Val, Size, &Unique, &Occur, NbUnique, equal);

  // Remove Nan from unique
  if(FisIsnan(Unique[NbUnique - 1])) NbUnique --;
}


void INHFP::KmeansInitPart(int n)
  //*****************************
{
  int i, empty, k;
  int * eff;
  double *Centres;
  double iter;

  if(n < 1) 
    {
      sprintf(ErrorMsg,  "KmeansInitPart-~InvalidNumberOfGroups~: %d~\n~", n);
      throw std::runtime_error( ErrorMsg );
    }

  Centres = new double [n];
  eff = new int [n];
 
  if(n == 1) 
    {
      Centres[0] = 0.;
      eff[0]=0;
    }
  else   
    for(i = 0; i < n ; i++) 
      {	
	eff[i] = 0;
	Centres[i] = (double) i/(n-1);
      }

  iter = Kmeans(Val, Size, Centres, n);
       	
  for(i = 0; i <Size;i++) eff[AssignClas(Val[i],Centres,n)]++;
      
  empty = 0;
  for(i = 0; i < n ; i++) 
    if (!eff[i]) empty++;
  NbUnique = n - empty;
  // printf("\n init using kmeans %d non empty groups\n", NbUnique);

  Unique = new double [NbUnique];
  Occur = new int [NbUnique];
  k=0;
  for(i = 0; i < n ; i++)
    if (eff[i])
      {
	Occur[k] = eff[i];
	Unique[k] = Centres[i];
	k++;
      }
  
  delete [] Centres;
  delete [] eff;
}


void INHFP::PrintCfgHfp(int num, FILE *f)
  //*************************************
{
  char Chaine[4];
  if(active) sprintf(Chaine, "yes");
  else sprintf(Chaine, "no");
      
  FISIN::PrintCfgTag(num, f);
  fprintf(f, "Active=%c%s%c\n", STRING_SEP, Chaine, STRING_SEP);
  fprintf(f, "Name=%c%s%c\n", STRING_SEP, Name, STRING_SEP);
  fprintf(f, "Range=%c%f%c%f%c\n", START_NB, ValInf, SEPARE, ValSup, END_NB);
  fprintf(f, "NMFs=%d\n", NmfMax);
}


void INHFP::PrintDist(int num) 
  //***************************
{
  int i, j;

  printf("Input %d : \n", num);
  for(i = 0; i < NbUnique; i++)
    {
      for(j = 0; j < NbUnique; j++)
	printf("%10.5f", Dist[i][j]);
      printf("\n");
    }
  printf("\n");
}


double INHFP::ExtDist(int a, int b)
  //*******************************
{
  if(b == a) return 0.0;
 
  CSeg an, bn;

  if(! Symbolic) // Distance numerique
    {
      Fp[a]->Kernel(an);
      Fp[b]->Kernel(bn);
    }

  if(b > a)   
    {
      if(Symbolic) return 0.5 + (double) (b - a) / (Nmf-1);
      else return bn.l - an.l + 0.5;   
    }
  
  if(b < a)   
    {
      if(Symbolic) return 0.5 + (double) (a - b) / (Nmf-1);
      else return an.l - bn.l + 0.5; 
    }

  return 0.0;
}


double INHFP::DistSum(int s, int before)
  //************************************
{
  int i, j, SefDebut, SefFin;
  double somme = 0.0;
  int Tous = false;

  if(before) Nmf++;

  if((s > -1) && (NmfDist == 2)) Init2MF(s, before, SefDebut, SefFin);   
  else if((s > -1) && (NmfDist == 3)) Init3MF(s, before, SefDebut, SefFin);
  else if((s > -1) && (NmfDist == 4)) Init4MF(s, before, SefDebut, SefFin);
  else 
    {
      Tous = true;
      SefDebut = 0;
      SefFin = Nmf-1;
    }

    if(Dsimpli == 1) SimpliDist(SefDebut, SefFin);
  else Distance(SefDebut, SefFin);

  if(before) Nmf--;

  for(i = 0; i < NbUnique; i++)
    for(j = i; j < NbUnique; j++)
      somme += Dist[i][j] * Occur[i] * Occur[j];

  if(Tous) somme /= Size * (Size - 1) * 2.;

  return somme;
}

void INHFP::PrintVertices(FILE * f)
  //*******************************
{
  int i, j;
  double diff;
  
  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;
  
  fprintf(f, "\n%s : %s  NmfInit : %d ", GetType(), Name, NmfInit);
  fprintf(f, "\n");
  for(i = 1; i < Nvertices; i++) 
    {
      fprintf(f, "%2d %5.2f %5.3f", i+1, Vertices[i][0].l,  Vertices[i][0].r); 
      
      for(j = 1; j < i + 2; j ++)  
	{
	  //fprintf(f, "   %3.2e %3.2e", Vertices[i][j].l * diff + ValInf, Vertices[i][j].r * diff + ValInf);

	  // Print only one value for triangular shaped sets
	  if(Vertices[i][j].p)
	    fprintf(f, "  %3.2e*", Vertices[i][j].l * diff + ValInf);
	  else
	    fprintf(f, "  %3.2e ", Vertices[i][j].l * diff + ValInf);
	}
      fprintf(f, "\n");
    } 
}


void INHFP::StoreVertices(FILE * f)
  //*******************************
{
  int i, j;
  double diff;
  
  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;
  
  fprintf(f, "%d\n", Nvertices);
  for(i = 1; i < Nvertices; i++) 
    {
      fprintf(f, "%f%c%f", Vertices[i][0].l,  SEPARE, Vertices[i][0].r); 
      
      for(j = 1; j < i + 2; j ++)  
	fprintf(f, "%c%f", SEPARE, Vertices[i][j].l * diff + ValInf);
      fprintf(f, "\n");
    } 
}

void INHFP::ReadVertices(ifstream *f, int bufsize)
  //**********************************************
{
  int i, j;
  double diff;
  char *buf;
  double *Tab;

  Tab = new double [5];
  buf = new char[bufsize];
  
  diff = ValSup - ValInf;
  if(diff < EPSILON) diff = 1.0;

  f->getline(buf, bufsize);
  Nvertices = atoi(buf);
  if(Nvertices == 0) return;
  if(Nvertices < 0) 
    {
      sprintf(ErrorMsg, "~UnexpectedNumberOfValues~ in the vertices file : %d~", Nvertices);
      delete [] Tab; delete [] buf;
      throw std::runtime_error( ErrorMsg );
    }    
  
  delete [] Tab;
  Tab = new double [Nvertices + 2];

  Vertices = new TVertex * [Nvertices];
  for(i = 0; i < Nvertices; i++) 
    Vertices[i] = new TVertex  [i + 2];

  for(i = 1; i < Nvertices; i++) 
    {
      f->getline(buf, bufsize);
      if((strlen(buf) == 0) || (buf[0] == 0x00D) || 
	 (SearchNb(buf, Tab, i+3, SEPARE) != i+3) )
	{
	  sprintf( ErrorMsg, "~UnexpectedNumberOfValues~ in the vertices file : %d values expected ", i+3);
	  delete [] Tab; delete [] buf;
	  throw std::runtime_error( ErrorMsg );
	}

      Vertices[i][0].l = Tab[0]; 
      Vertices[i][0].r = Tab[1]; 

      for(j = 1; j < i + 2; j ++)  
	Vertices[i][j].l = Vertices[i][j].r = (Tab[j+1] - ValInf) / diff ;
    } 
  
  delete [] buf;
  delete []Tab;
}

int INHFP::HRegular(void)
  //*********************
{
  int i, j;
 
  for(j = 1; j < Nvertices; j++)
    {
      Vertices[j][0].l = Vertices[j][0].r = 0.;
      
      for(i = 1; i < (j+2); i++)
	Vertices[j][i].l = Vertices[j][i].r = (double) (i-1)/j;
    }
  return 0;
}
  
int INHFP::HKmeans(void)
  //********************
{
  int i, j, iter;
  double *Centres;

  Centres = new double [Nvertices];

  for(j = 1; j < Nvertices; j++)
    {
      Vertices[j][0].l = Vertices[j][0].r = 0.;
       
      for(i = 0; i < j+1; i++) 	Centres[i] = (double) i/j;

      iter = Kmeans(Val, Size, Centres, j+1);
 	
      for(i = 1; i < (j+2); i++)
	Vertices[j][i].l = Vertices[j][i].r = Centres[i-1];
    }

  delete [] Centres;
  return 0;
}

//*************************************************************************
//
//
//                           PROTECTED FUNCTIONS
//
//
//*************************************************************************
int INHFP::InitNmfNvertices(void)
  //*****************************
{
  int j; 

  NmfMax = 0;
  if(Nvertices == 0) active = false;
  if(active == false) return 0;
   
  NmfMax = Nvertices;

  Fp = new MF * [NmfMax];
  for(j = 0; j < NmfMax; j++) Fp[j] = NULL;
  return NmfMax;
}

int INHFP::SetNmf(void)
  //*******************
{
  int j, nb; 

  if(active == false) return 0;
   
  if(NmfMax > Nvertices)
    {
      //      sprintf(ErrorMsg, "\nNumber of MF requested (%d) greater than number of vertices (%d) for input %s",  NmfMax, Nvertices, Name);
      //      throw std::runtime_error( ErrorMsg );
      nb = Nvertices;
      printf("\nNumber of MF requested (%d) greater than number of vertices (%d) for input %s",  NmfMax, Nvertices, Name);
    }
  else  nb = NmfMax;

  Fp = new MF * [nb];
  for(j = 0; j < nb; j++) Fp[j] = NULL;
  return nb;
}

void INHFP::Init2MF(int s, int before, int & mfmin, int & mfmax)
  //************************************************************
{
  mfmin = s;

  if(before) mfmax = s+1;
  else mfmax = s;

} 

void INHFP::Init3MF(int s, int before, int & mfmin, int & mfmax)
  //************************************************************
  // Take into account 3 of the 4 modified MF: 
  // the 3 first ones for the first part of the partition,
  // the 3 last ones for the other.
  // Objectif: remove side effects (effets de bord).
{
  if(before) // 3 mf
    {
      if(s < (Nmf - 1) / 2)
	{
	  mfmin = s;
	  mfmax = s + 2;
	}
      else 
	{
	  mfmin = s - 1;
	  mfmax = s + 1;
	}
    }
  else       // 2 mf
    {
      if(s < (Nmf - 1) / 2)
	{
	  mfmin = s;
	  mfmax = s + 1;
	}
      else 
	{
	  mfmin = s - 1;
	  mfmax = s;
	}
    }
} // Fin de Init3MF

void INHFP::Init4MF(int s, int before, int & mfmin, int & mfmax)
  //************************************************************
  // Consider all the modified MF, at most 4 before the merging, and 3 after
{
  if(s) mfmin = s-1; 
  else mfmin = s;

  if(before)  // 4 mf max
    {
      if(s < Nmf - 2) mfmax = s+2;
      else mfmax = s + 1;
    }
  else       // 3 mf max
    {
      if(s < Nmf - 1) mfmax = s+1;
      else mfmax = s;
    }

  if (mfmin == s) mfmax=mfmax+1;
  if (before) {if (s >= Nmf-2) mfmin=mfmin-1;}
  else {if (s >= Nmf-1) mfmin=mfmin-1;}
} 


void INHFP::SimpliDist(int smin, int smax)
  //**************************************
{
  int i, j, k, max_i1, max_i2, max_j1, max_j2;
  double mu_i1, mu_i2, mu_j1, mu_j2;
  double mu;
  double Dc=0.5;

  for(i = 0; i < NbUnique; i++)
    {
      Dist[i][i] = 0.0;
      mu_i1 = mu_i2 = 0.;
      max_i1 = max_i2 = INACTIVE;
      GetDegsV(Unique[i]); 

      for(k = smin; k <= smax; k++)
	{
	  mu=Mfdeg()[k];
	  if(mu > EPSILON) 
	    {
	      if(max_i1 == INACTIVE) 
		{ 
		  mu_i1 = mu;
		  max_i1 = k;
		}
	      else
		{
		  mu_i2 = mu;
		  max_i2 = k;
		}
	    } 
	}

      if (max_i2 == INACTIVE && max_i1 != INACTIVE)
	{
	  if(max_i1 == smin && smin > 0 && Mfdeg()[smin-1] > EPSILON)
	    {
	      max_i2 = smin;
	      mu_i2 = mu_i1;
	      max_i1 = smin-1;
	      mu_i1 = Mfdeg()[smin-1];
	    }
	  else if(max_i1 == smax && smax < (Nmf-1) && Mfdeg()[smax+1] > EPSILON)
	    {
	      max_i2 = smax+1;
	      mu_i2 = Mfdeg()[smax+1];
	    }
	}
      
      for(j = i + 1; j < NbUnique; j++) 
	{
	  Dist[i][j] = Dist[j][i] = 0.;
	  mu_j1 = mu_j2 = 0.;
	  max_j1 = max_j2 = INACTIVE;
	  GetDegsV(Unique[j]); 

	  for(k = smin; k <=smax ; k++)
	    {
	      mu = Mfdeg()[k];
	      if(mu > EPSILON) 
		{
		  if(max_j1 == INACTIVE)
		    { 
		      mu_j1 = mu;
		      max_j1 = k;
		    }
		  else
		    {
		      mu_j2 = mu;
		      max_j2 = k;
		    }
		} 
	    } 

	  if (max_j2 == INACTIVE && max_j1 != INACTIVE)
	    {
	      if(max_j1 == smin && smin > 0 && Mfdeg()[smin-1] > EPSILON)
		{
 		  max_j2 = smin;
		  mu_j2 = mu_j1;
		  max_j1 = smin-1;
		  mu_j1 = Mfdeg()[smin-1];
		}
	      else if(max_j1 == smax && smax < (Nmf-1) && Mfdeg()[smax+1] > EPSILON)
		{
		  max_j2 = smax+1;
		  mu_j2 = Mfdeg()[smax+1];
		}
	    }
      

	  // protection : un des 2 points i,j n'appartient a aucun des sef de [smin,smax]
	  if((max_i1 == INACTIVE) || (max_j1 == INACTIVE)) continue;

	  // Premier cas : appartenance unique pour chacun
	  if(max_i2 == INACTIVE && max_j2 == INACTIVE)
	    {
	      Dist[i][j] = IntDist(mu_i1, mu_j1); 
	  
	      if(max_i1 != max_j1) // distance externe
		Dist[i][j] += ExtDist(max_i1, max_j1);
	    }
	    

	  // Deuxieme cas : Un des points a une appartenance double
	  // On fait l'hypothese que les mu sont voisins
	  else if(max_i2 == INACTIVE && max_j2 != INACTIVE) // C'est j.
	    {
	      // Dist[i][j] = IntDist(mu_i1, mu_j1)/2 + IntDist(mu_i1, mu_j2)/2 + 
	      //   ExtDist(max_i1, max_j2)/2 + ExtDist(max_i1, max_j1)/2 si necessaire;
	      // Mu_j1 voisin de mu_j2 donc les deux distances internes sont voisines :
	      Dist[i][j] = IntDist(1., mu_j1*2.) + ExtDist(max_i1, max_j2)/2;
	      if(max_j1 != max_i1)
		Dist[i][j] += ExtDist(max_i1, max_j1)/2;
	    }

	  else if(max_j2 == INACTIVE && max_i2 != INACTIVE) // C'est i.
	    {
	      Dist[i][j] = IntDist(1., mu_j1*2.) + ExtDist(max_i1, max_j1)/2;
	      if(max_i2 != max_j1)
		Dist[i][j] += ExtDist(max_i2, max_j1)/2 ;
	    }
	 
	  // Troisieme cas : les deux ont des appartenances doubles
	  else if(max_i2 != INACTIVE && max_j2 != INACTIVE) 
	    {
	      if(max_i1 == max_j1) 
		// Les deux points appartiennent aux memes ensembles
		Dist[i][j] = IntDist(mu_i1, mu_j1) + ExtDist(max_i1, max_j2)/2;
	      else if(max_i2 == max_j1) 
		// Ils ont un et un seul ensemble en commun
		Dist[i][j] = (3.*IntDist(mu_i1, mu_j2)/4.) + ExtDist(max_i1, max_j2)/2 + Dc/4.;
	      else
		// Ils n'ont aucun ensemble en commun
		Dist[i][j] = 0.5*ExtDist(max_i1, max_j1) + 0.5*ExtDist(max_i2, max_j2) + 0.5*IntDist(mu_i1, mu_j1) + 0.5*IntDist(mu_i2, mu_j2);
	    }

	  Dist[j][i] = Dist[i][j];
	}
    }
} 

void INHFP::Distance(int smin, int smax)
  //************************************
{
  int i, j, k, ki, kj, OKi, OKj;
  double *VaI, SumI, SumJ, Div;

  VaI = new double [Nmf];

  for(i = 0; i < NbUnique; i++)
    {
      Dist[i][i] = 0.0;
      GetDegsV(Unique[i]); 
      OKi = false;

      for(k = 0; k < Nmf ; k++) 
	VaI[k] = Mfdeg()[k];

      for(k = smin; k <= smax ; k++) 
	if(VaI[k] > EPSILON) OKi = true;
      // Point i has at least a non null degree to one of the MF of interest

      for(j = i + 1; j < NbUnique; j++) 
	{
	  Dist[i][j] = Dist[j][i] = 0.;
	  if(OKi == false) continue;

	  GetDegsV(Unique[j]); 
	  OKj = false;
      

	  for(k = smin; k <= smax ; k++) 
	    if(Mfdeg()[k] > EPSILON) OKj = true;

	  if(OKj == false) continue;

	  if(Symbolic > -1) // Old distance with two options symbolic and numerical
	    {
	      // All the degrees are taken into account
	      SumI = 0.; SumJ = 0.;
	      
	      for(ki = 0; ki < Nmf ; ki++)
		{
		  if(VaI[ki] < EPSILON) continue;
		  SumI += VaI[ki];
		  SumJ = 0.;
		  for(kj = 0; kj < Nmf ; kj++)
		    {
		      if(Mfdeg()[kj] < EPSILON) continue;
		      SumJ += Mfdeg()[kj];
		      
		      Dist[i][j] += VaI[ki] * Mfdeg()[kj] * IntDist(VaI[ki], Mfdeg()[kj]);
		      if(ki != kj)
			Dist[i][j] += VaI[ki] * Mfdeg()[kj] * ExtDist(ki, kj);
		    }
		}
	      Div = SumI * SumJ;
	      if(Div) Dist[i][j] /= Div;
	      Dist[j][i] = Dist[i][j];
	    }
	  else Dist[j][i] = Dist[i][j] = FISIN::Distance(Unique[i], Unique[j]);
	}
    }

  delete [] VaI;
} 

double INHFP::Variance(double *Y,double *VarSef)
  //********************************************
{
  int i, k;
  double *Spi, *Spiyi, *Spiyi2, pk;    
  // Sums of weights, sum of y and sum of y^2
  double sum; // Return value 

  Spi = new double [Nmf]; 
  Spiyi = new double [Nmf]; 
  Spiyi2 = new double [Nmf]; 

  for(k = 0; k < Nmf; k++) Spi[k] = Spiyi[k] =  Spiyi2[k] = 0.0;

  for(i = 0; i < Size; i++)
    {
      GetDegsV(Val[i]); 

      for(k = 0; k < Nmf; k++)
	{
	  if((pk = Mfdeg()[k]) > EPSILON)
	    {
	      Spi[k] += pk;
	      Spiyi[k] += pk * Y[i];
	      Spiyi2[k] += pk * Y[i] * Y[i];
	    }
	}
    }
  
  sum = 0.0;

  for(k = 0; k < Nmf; k++)
    {      
      VarSef[k] = Spiyi2[k] - (Spiyi[k] * Spiyi[k]) / Spi[k];
      sum += VarSef[k];
      VarSef[k] /= Spi[k];
    }

  delete [] Spi;
  delete [] Spiyi;
  delete [] Spiyi2;

  if (sum > EPSILON) return sum ;
  else return 0.0;
}

void INHFP::VarianceNP(double *Y, int e, double &v, double &p)
  //**********************************************************
{
  int i;
  double Spiyi, Spiyi2, pk;    

  p = Spiyi = Spiyi2 = 0.0;

  for(i = 0; i < Size; i++)
    {
      pk = GetADeg(e, Val[i]);

      if(pk > EPSILON)
	{
	  p += pk;
	  Spiyi += pk * Y[i];
	  Spiyi2 += pk * Y[i] * Y[i];
	}
    }
  v = (Spiyi2 - (Spiyi * Spiyi) / p) / p;
  if(v < EPSILON) v = 0.0;
  return ;
}

void INHFP::Density(double * Dens)
  //******************************
{
  int i, k;
  double *Card, *Aire;
  CSeg c, p, s;

  Card = new double [Nmf];
  Aire = new double  [Nmf];

  for(i = 0; i < Nmf; i++)
    Card[i] = Aire[i] = 0.0;

  // Fuzzy cardinality computation
  for(i = 0; i < NbUnique; i++)
    {
      GetDegsV(Unique[i]); 
     
      for(k = 0; k < Nmf; k++)
	if(Mfdeg()[k] >= EPSILON)  
	  Card[k] += Mfdeg()[k];
    }    

  // Area computation
  // Aire d'1 sef triangulaire normalise : support / 2
  // Aire commune aux sef i et i+1 du fait de la partition floue forte sommet(i+1) - sommet(i) /4

  for(k = 0; k < Nmf; k++)
    {
      Fp[k]->Support(c);
      //protection si les triangles extremes vont a l'inendi
      //cas des demi-trapezes, gestion analogue a FIS::ReadMf
      // 0. et 1. correspondent aux bornes du domaine de l'entree
      // (normalisees)
		
      if (c.l==-INFINI_HFP) c.l=0.;
      if (c.r==INFINI_HFP) c.r=1.;

      Aire[k] = (c.r - c.l) / 2;
     
	
      if(k)                    // Enlever aire commune avec precedent
	{
	  Fp[k-1]->Kernel(p);
	  Fp[k]->Kernel(c);
	  Aire[k] -= (c.l - p.l) / 4;
	 
	}
      
      if(k < Nmf - 1)        // Enlever aire commune avec suivant
	{
	  Fp[k+1]->Kernel(s);
	  Fp[k]->Kernel(c);
	  Aire[k] -= (s.l - c.l) / 4;
	}
    }

  for(k = 0; k < Nmf; k++) 
    {
      if(Aire[k] < EPSILON) Aire[k] = 1.0;
      Dens[k] = Card[k] / Aire[k];
    }

  delete [] Card;
  delete [] Aire;
} 

double INHFP::EvalMass(int f)
  //*************************
{
  int i;
  double masse =  0.0;


  for(i = 0; i < NbUnique; i++)
    masse += Fp[f]->GetDeg(Unique[i]) * Occur[i];

  return masse;
}


void INHFP::InMerge(double * Ys)
  //*****************************
  // Examine l'ensemble des possibilites de fusion et memorise la meilleure
  // combinaison a chaque etape
  // Ys : tableau des values de sortie
{
  if(Ys == NULL && VarCrit)
    {
      VarCrit = 0; // distance
      printf("\nImpossible d'utiliser le critere de variance pour la fusion sans sortie.");
      printf("\nUtilisation du critere de distance.\n");
    }

  if(Nmf == Nvertices)
    {
      int i;
      double *T, med, moy, delta, Tmin, Tmax;
      T = new double [Nmf];

      for (i = 0; i < Nmf; i++) T[i] = 0.0;

      Density(T);
      StatArray(T, Nmf, 0, med, moy, delta, Tmin, Tmax, 0);
      Vertices[Nmf-1][0].l = delta;
      // Vertices[Nmf-1][0].l = DistSum(-1, 0);
      // Nmf est correct, -1 force l'appel de DistanceTous
      for (i = 0; i < Nmf; i++) T[i] = 0.0;
      if(Ys) Vertices[Nmf-1][0].r = Variance(Ys, T);
      else Vertices[Nmf-1][0].r=0.0;

      for(i = 0; i < Nmf; i++)
	{
	  Fp[i]->Kernel(Vertices[Nmf-1][i+1].l, Vertices[Nmf-1][i+1].r);
	  Vertices[Nmf-1][i+1].p = 0;
	  if(Vertices[Nmf-1][0].r && T[i] < VarThres) Vertices[Nmf-1][i+1].p = 1;
	}
      
      delete [] T;
    }
  
  SelectMf(Ys);

}    
//*****************        End of HFPIN.CPP    *********************************

